"""
Sage package for using Shimura's reciprocity law for Siegel modular funcions.

Version 0.1 -- 17th July 2012 -- tested with Sage 5.1

When using this package in a publication, please cite

- Marco Streng -- An explicit version of Shimura's reciprocity law for Siegel
  modular functions. arXiv:1201.0020 or preferably a future published version.

This package is based on formulas from that paper, is motivated by it, and
contains examples from it.

#***************************************************************************** 
#       Copyright (C) 2010 -- 2012 Marco Streng <marco.streng@gmail.com>
# 
#  Distributed under the terms of the GNU General Public License (GPL) 
# 
#    This code is distributed in the hope that it will be useful, 
#    but WITHOUT ANY WARRANTY; without even the implied warranty of 
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#    General Public License for more details. 
# 
#  The full text of the GPL is available at: 
# 
#                  http://www.gnu.org/licenses/ 
#***************************************************************************** 

To use this package, start Sage with the .sage files from this package in your
working directory. Then type::

    sage: load("main.sage")

See below for some examples.

The current version of this package is suitable when working with theta
characteristics in (1/den)ZZ and g-dimensional abelian varieties where
den^(2*g) is small and den is even (e.g. g=den=2). This already allows for
class invariants that yield a significant improvement in size relative to Igusa
invariants.


EXAMPLES:

We start with the first example presented at Geocrypt and ECC::

    sage: load("main.sage")
    sage: K = CM_Field([521,27,52])
    sage: Z = K.one_period_matrix('Phi'); Z
    Period Matrix
    [-0.38032? + 0.92486?*I  0.42520? + 0.38513?*I]
    [ 0.42520? + 0.38513?*I    0.2607? + 1.3683?*I]
    sage: I = igusa_invariants_absolute()
    sage: Kr0 = QuadraticField(13,'a')
    sage: gammas = reciprocity_map_image(Z, 8)
    sage: c = Z.complex_conjugation_symplectic_matrix(8)
    sage: visualize(gammas + [c], 2) # random
    On the 8-th powers, the action is (note that 16 means 0):
    1: (1,6,16)(2,3,4)(8,15,9)
    2: (1,6,16)(2,3,4)(8,15,9)
    3: ()
    4: ()
    5: ()
    6: ()
    7: (1,4)(2,16)(3,6)(8,15)
    Permutation Group with generators [(), (1,4)(2,16)(3,6)(8,15), (1,6,16)(2,3,4)(8,15,9)] of order 6
    The action on the orbit [0, 1, 2, 3, 4, 6] is as follows
    [            0|            1             2             3             4             5             6             7]
    [-------------+-------------------------------------------------------------------------------------------------]
    [           t0|(-zeta8^2)*t1 (-zeta8^2)*t1            t0            t0            t0            t0            t2]
    [           t1| (zeta8^3)*t6  (zeta8^3)*t6           -t1            t1           -t1            t1            t4]
    [           t2|          -t3           -t3           -t2            t2           -t2            t2            t0]
    [           t3|   (zeta8)*t4    (zeta8)*t4            t3            t3            t3            t3            t6]
    [           t4|(-zeta8^2)*t2 (-zeta8^2)*t2            t4            t4            t4            t4            t1]
    [           t6|           t0            t0           -t6            t6           -t6            t6            t3]
    [      (zeta8)|   (-zeta8^3)    (-zeta8^3)      (-zeta8)       (zeta8)      (-zeta8)       (zeta8)    (-zeta8^3)]
    The action on the orbit [8, 9, 15] is as follows
    [             0|             1              2              3              4              5              6              7]
    [--------------+--------------------------------------------------------------------------------------------------------]
    [            t8|   (zeta8)*t15    (zeta8)*t15            -t8             t8            -t8             t8 (-zeta8^3)*t15]
    [            t9| (-zeta8^2)*t8  (-zeta8^2)*t8             t9             t9             t9             t9     (zeta8)*t9]
    [           t15| (-zeta8^2)*t9  (-zeta8^2)*t9           -t15            t15           -t15            t15  (-zeta8^3)*t8]
    [       (zeta8)|    (-zeta8^3)     (-zeta8^3)       (-zeta8)        (zeta8)       (-zeta8)        (zeta8)     (-zeta8^3)]
    The action on the orbit [12] is as follows
    [             0|             1              2              3              4              5              6              7]
    [--------------+--------------------------------------------------------------------------------------------------------]
    [           t12|   (zeta8)*t12   (-zeta8)*t12            t12           -t12           -t12           -t12 (-zeta8^3)*t12]
    [       (zeta8)|    (-zeta8^3)     (-zeta8^3)       (-zeta8)        (zeta8)       (-zeta8)        (zeta8)     (-zeta8^3)]
    [t12^3/(t8*t9*t15), zeta8*t12^3/(t8*t9*t15), i*t12^3/(t8*t9*t15), zeta8^3*t12^3/(t8*t9*t15)]
    sage: R = theta_ring(2,2)[0]
    sage: zeta8 = R.base_ring().gen()    
    sage: T = [ThetaModForm(zeta8^j*R.gens()[12]^3 / prod([R.gens()[i] for i in [8,9,15]])) for j in range(4)]
    sage: [len(t.orbit(gammas+[c])) for t in T] # long time
    [4, 2, 4, 2]
    sage: i = T[1]^2; i
    (zeta8^2)*t12^6/(t8^2*t9^2*t15^2)

In this case, the class number is odd and the class number of
the real quadratic field is 1. This implies that the field of moduli
is the Hilbert class field, and we can use the class group.::

    sage: shrecip = [Z.Shimura_reciprocity(A.ideal(), n=8, m=1, transformation=True, period_matrix=True) for A in Z.CM_type().reflex_field().class_group()]
    sage: d = len(shrecip)
    sage: funcs = [[i**u for (U,M,u) in shrecip]] + [[j for k in range(d)] for j in I]
    sage: rts = [[funcs[j][k](shrecip[k][0], prec=1000) for k in range(d)] for j in range(4)] # long time almost 4 minutes
    sage: X = PolynomialRing(rts[0][0].parent(), 'X').gen() # long time (not really, the only thing that takes time is the previous one)
    sage: minpols_num = [prod([X-r for r in rts[k]]) for k in [0,1]] # long time
    sage: minpols = [recognize_polynomial(p, Kr0, N=None, emb=None) for p in minpols_num] # long time
    sage: intpols_num = [[short_interpolation(rts[k], rts[j]) for j in range(k+1,4)] for k in [0,1]] # long time
    sage: intpols = [[recognize_polynomial(p, Kr0, N=None, emb=None) for p in s] for s in intpols_num] # long time
    sage: without_clinv = [minpols[1]] + intpols[1]; without_clinv # long time
    [y^7 + (-155205162116358647755/10201*a + 559600170220938887110/10201)*y^6 + (-152407687697460195175920750535594152550/10201*a + 549513732768094956258970636118192859400/10201)*y^5 + (-2201909580030523730272623848434538048317834513875/20402*a + 7939097894735431844153019089320973153011210882125/20402)*y^4 + (-1047175262927393182849164587480891367594710449395570625/10201*a + 3775644104882200832865729346429752069380200097845736875/10201)*y^3 + (-907392914800494855136752991106041311116404713247380607234375/20402*a + 3271651681305911192688931423723753094763461200379169938284375/20402)*y^2 + (-15014166049656519860045880222971244113390650525905069987454062500/10201*a + 54134345550367190785605984445586939893083531851405365978411062500/10201)*y - 320854170291151322128777010521751890513120770505490537777676328984375/20402*a + 1156856162931200670387093211443242850125709667683265459917987279296875/20402, (155942160719197448511497600/104060401*a - 562257456480820026589520000/104060401)*y^6 + (10915460249997911281051048769982462340888000/104060401*a - 39356251626656444452197645346830542588488000/104060401)*y^5 + (168373146277549827762748743327083206237502305644120000/104060401*a - 607078012314904622487588715272472722561309748288920000/104060401)*y^4 + (238652435808138594034697534364809573290025398381044081800000/104060401*a - 860473594320621938090309645042531340247319597576659017800000/104060401)*y^3 + (104322262281490071026402121264030948196570701298335683621237000000/104060401*a - 376139265828315347221671304362880264396213048032027706818473000000/104060401)*y^2 + (3422978757984824090435381776507567613874776553028721788254983445000000/104060401*a - 12341725426738324424199494569900641042064837165414213925317002945000000/104060401)*y + 25444851830157171979850471655958457967719020294854199175794590509050000000/104060401*a - 91742717970215413696542092180092916555240944945393570493326703047900000000/104060401, (-4012937435827235689317264963498305932800000/104060401*a + 14468851690104080323524823696416410496000000/104060401)*y^6 + (-1506913225655983606143240710922345533620775435750856464000000/104060401*a + 5433252902777486004872984777262721832123795770308636080000000/104060401)*y^5 + (-10885562196559519508593359565510553063925803986258652826001713920000000/104060401*a + 39248452661947760486383130049374702797762681330867052835728546240000000/104060401)*y^4 + (-10353823389206156431412892798015311801020716070867604781938124565025200000000/104060401*a + 37331241126881141754569229980807920080905304920194122898918893817217200000000/104060401)*y^3 + (-4485870855187365031380849955017652628982072100341118267226436318108114676000000000/104060401*a + 16174037383487540399908140658405308702506346444494976170181843406011298948000000000/104060401)*y^2 + (-148450817277784374908896669100295472510110826966451685368788333734587930637950000000000/104060401*a + 535247033579587071899540245431137800567789741426378635706026376130489367192830000000000/104060401)*y - 1586204110477319767185402578315040039982207174583408730603126895573612283837343100000000000/104060401*a + 5719140253677722869121206183461376151381382068131288660643535348602757766642039900000000000/104060401]
    sage: with_clinv = [minpols[0]] + intpols[0]; with_clinv # long time
    [y^7 + (30056912/91809*a - 105080560/91809)*y^6 + (-2510103169664/826281*a + 9050615361664/826281)*y^5 + (-31191346986373120/7436529*a + 112462009885048832/7436529)*y^4 + (-2349120383562514432/66928761*a + 8469874588158623744/66928761)*y^3 + (-26197067707249590272/22309587*a + 94454871140377034752/22309587)*y^2 + (250917334141632512/66928761*a - 904696010264018944/66928761)*y - 499961036800/91809*a + 1800799256576/91809, (155205162116358647755/10201*a - 559600170220938887110/10201)*y^6 + (4293512808155549781824000/10201*a - 15480480581665978617394880/10201)*y^5 + (4796992628983857601067816320/10201*a - 17295802891824100262459263360/10201)*y^4 + (120945583333941400455909591040/30603*a - 436075502251428602426380912640/30603)*y^3 + (1341042908932829083414600581120/10201*a - 4835198970754700340106176593920/10201)*y^2 + (-38533828940561016639837962240/91809*a + 138935696085150872191306301440/91809)*y + 167832146481204715187077120/275427*a - 605127409809396328300544000/275427, (155942160719197448511497600/104060401*a - 562257456480820026589520000/104060401)*y^6 + (-166301726476865527526007193600/312181203*a + 599609402010406725455728179200/312181203)*y^5 + (-2476327951737479042110141713203200/936543609*a + 8928527404854192205498136257740800/936543609)*y^4 + (-552147133863772061563401183356518400/25286677443*a + 1990794802746309324199714339363225600/25286677443)*y^3 + (-56403828236193801907781802283289804800/75860032329*a + 203366894838060336544487139786201497600/75860032329)*y^2 + (20008870365250292462430263758028800/8428892481*a - 72143008066021761377656153466470400/8428892481)*y - 784327759750570090436933294489600/227580096987*a + 2827933954549445333497544074854400/227580096987, (-4012937435827235689317264963498305932800000/104060401*a + 14468851690104080323524823696416410496000000/104060401)*y^6 + (-16711094246362976345862986728316071181619200000/104060401*a + 60252707174372962283239786551247279979110400000/104060401)*y^5 + (-14069073960477216641209061802589478970032128000000/104060401*a + 50726767562795827750926416934161464151375872000000/104060401)*y^4 + (-119107880103422386089716488866303676654026752000000/104060401*a + 429449569024706497123435462920155844683995545600000/104060401)*y^3 + (-11770178875083208903917412174969509904249782272000000/312181203*a + 42437983455495566570747427020049615148124471296000000/312181203)*y^2 + (12526191764620323384312023796102221506858188800000/104060401*a - 45163826693633325731560527040984621868751257600000/104060401)*y - 163671667363475421464773039962382760987852800000/936543609*a + 590126589019696595537103376578535359512576000000/936543609]
    sage: [(p*101^2*2^28*3^8).denominator() for p in minpols] # long time
    [1, 1]
    sage: [[(p*3^7*101^4).denominator() for p in s] for s in intpols] # long time
    [[1, 1, 1], [1, 1]]
    sage: X = minpols[0].parent().gen() # long time
    sage: minpols[0](2^3*X)/2^21 # long time
    y^7 + (3757114/91809*a - 13135070/91809)*y^6 + (-39220362026/826281*a + 141415865026/826281)*y^5 + (-60920599582760/7436529*a + 219652363056736/7436529)*y^4 + (-573515718643192/66928761*a + 2067840475624664/66928761)*y^3 + (-799471060401904/22309587*a + 2882533909313264/22309587)*y^2 + (957173668448/66928761*a - 3451141396576/66928761)*y - 238400/91809*a + 858688/91809

    
We continue with the second example presented at the same conferences::

    sage: K = CM_Field([709, 310, 17644])
    sage: O = K.maximal_order()
    sage: Phi = K.CM_types()[3]
    sage: d = K.different().gens_reduced()[0]
    sage: xi = d^-1
    sage: [phi(xi) for phi in Phi]
    [0.00122524459004800?*I, 0.00216656865333580?*I]
    sage: Z = PeriodMatrix(Phi, 1*O, xi)
    sage: gammas = reciprocity_map_image(Z, 8) # long time, less than 7 minutes, note: requires sage-4.7.1 or higher, see trac ticket 11234
    sage: c = Z.complex_conjugation_symplectic_matrix(8) # long time
    sage: visualize(gammas + [c], 2) # long time
    On the 8-th powers, the action is (note that 16 means 0):
    1: ()
    2: ()
    3: ()
    4: ()
    5: ()
    6: ()
    7: ()
    8: ()
    9: ()
    Permutation Group with generators [()] of order 1
    The action on the orbit [1] is as follows
    [           0|           1            2            3            4            5            6            7            8            9]
    [------------+--------------------------------------------------------------------------------------------------------------------]
    [          t1|          t1           t1 (zeta8^2)*t1           t1           t1           t1           t1           t1           t1]
    [     (zeta8)|    (-zeta8)      (zeta8)      (zeta8)      (zeta8)     (-zeta8)     (-zeta8)      (zeta8)      (zeta8)   (-zeta8^3)]
    The action on the orbit [2] is as follows
    [            0|            1             2             3             4             5             6             7             8             9]
    [-------------+-----------------------------------------------------------------------------------------------------------------------------]
    [           t2|           t2 (-zeta8^2)*t2           -t2            t2            t2            t2            t2            t2            t2]
    [      (zeta8)|     (-zeta8)       (zeta8)       (zeta8)       (zeta8)      (-zeta8)      (-zeta8)       (zeta8)       (zeta8)    (-zeta8^3)]
    The action on the orbit [3] is as follows
    [           0|           1            2            3            4            5            6            7            8            9]
    [------------+--------------------------------------------------------------------------------------------------------------------]
    [          t3|          t3 (zeta8^2)*t3 (zeta8^2)*t3           t3           t3           t3           t3           t3           t3]
    [     (zeta8)|    (-zeta8)      (zeta8)      (zeta8)      (zeta8)     (-zeta8)     (-zeta8)      (zeta8)      (zeta8)   (-zeta8^3)]
    The action on the orbit [4] is as follows
    [         0|         1          2          3          4          5          6          7          8          9]
    [----------+--------------------------------------------------------------------------------------------------]
    [        t4|        t4        -t4         t4         t4         t4         t4         t4         t4         t4]
    [   (zeta8)|  (-zeta8)    (zeta8)    (zeta8)    (zeta8)   (-zeta8)   (-zeta8)    (zeta8)    (zeta8) (-zeta8^3)]
    The action on the orbit [6] is as follows
    [           0|           1            2            3            4            5            6            7            8            9]
    [------------+--------------------------------------------------------------------------------------------------------------------]
    [          t6|         -t6 (zeta8^2)*t6          -t6           t6          -t6          -t6           t6           t6           t6]
    [     (zeta8)|    (-zeta8)      (zeta8)      (zeta8)      (zeta8)     (-zeta8)     (-zeta8)      (zeta8)      (zeta8)   (-zeta8^3)]
    The action on the orbit [8] is as follows
    [         0|         1          2          3          4          5          6          7          8          9]
    [----------+--------------------------------------------------------------------------------------------------]
    [        t8|        t8        -t8        -t8         t8         t8         t8         t8         t8         t8]
    [   (zeta8)|  (-zeta8)    (zeta8)    (zeta8)    (zeta8)   (-zeta8)   (-zeta8)    (zeta8)    (zeta8) (-zeta8^3)]
    The action on the orbit [9] is as follows
    [            0|            1             2             3             4             5             6             7             8             9]
    [-------------+-----------------------------------------------------------------------------------------------------------------------------]
    [           t9|          -t9           -t9 (-zeta8^2)*t9            t9           -t9           -t9            t9            t9            t9]
    [      (zeta8)|     (-zeta8)       (zeta8)       (zeta8)       (zeta8)      (-zeta8)      (-zeta8)       (zeta8)       (zeta8)    (-zeta8^3)]
    The action on the orbit [12] is as follows
    [         0|         1          2          3          4          5          6          7          8          9]
    [----------+--------------------------------------------------------------------------------------------------]
    [       t12|       t12        t12       -t12        t12        t12        t12        t12        t12        t12]
    [   (zeta8)|  (-zeta8)    (zeta8)    (zeta8)    (zeta8)   (-zeta8)   (-zeta8)    (zeta8)    (zeta8) (-zeta8^3)]
    The action on the orbit [15] is as follows
    [             0|             1              2              3              4              5              6              7              8              9]
    [--------------+--------------------------------------------------------------------------------------------------------------------------------------]
    [           t15|          -t15  (zeta8^2)*t15 (-zeta8^2)*t15            t15           -t15           -t15            t15            t15            t15]
    [       (zeta8)|      (-zeta8)        (zeta8)        (zeta8)        (zeta8)       (-zeta8)       (-zeta8)        (zeta8)        (zeta8)     (-zeta8^3)]
    The action on the orbit [0] is as follows
    [         0|         1          2          3          4          5          6          7          8          9]
    [----------+--------------------------------------------------------------------------------------------------]
    [        t0|        t0         t0         t0         t0         t0         t0         t0         t0         t0]
    [   (zeta8)|  (-zeta8)    (zeta8)    (zeta8)    (zeta8)   (-zeta8)   (-zeta8)    (zeta8)    (zeta8) (-zeta8^3)]
    
The prime 2 splits sufficiently much that Rosenhain invariants should always work.::

    sage: [e.orbit(gammas+[c]) == [e] for e in rosenhain_invariants(2)] # long time
    [True, True, True]
    
Here's a nice other small invariant::

    sage: R = theta_ring(2,2)[0]
    sage: zeta8 = R.base_ring().gen()    
    sage: ThetaModForm(R.gens()[4]/R.gens()[12]).orbit(gammas) # long time
    [t4/t12, (-t4)/t12]
    
Now the invariants of the talk at Geocrypt and ECC::

    sage: t = ThetaModForm(R.gens()[0]*R.gens()[8]/R.gens()[4]/R.gens()[12])
    sage: U = ThetaModForm(R.gens()[2]/R.gens()[6])^2
    sage: V = ThetaModForm(R.gens()[8]/R.gens()[12])^2
    sage: W = ThetaModForm(R.gens()[0]/R.gens()[4])^2
    sage: [a.orbit(gammas+[c]) for a in [t,U,V,W]] # long time
    [[t0*t8/(t4*t12)], [t2^2/t6^2], [t8^2/t12^2], [t0^2/t4^2]]
    sage: u = U*V
    sage: v = U*W
    sage: t^2/V*U == v
    True
    sage: [t,u,v] # these were presented first
    [t0*t8/(t4*t12), t2^2*t8^2/(t6^2*t12^2), t0^2*t2^2/(t4^2*t6^2)]
    sage: [t,U,V] # maybe these are smaller?
    [t0*t8/(t4*t12), t2^2/t6^2, t8^2/t12^2]
   
"""

load_attach_mode(load_debug=True)

load "basic.sage"
load "polynomials.sage"
load "symplectic_matrices.sage"
load "period_matrices.sage"
load "theta_trans.sage"
load "cm_types.sage"
load "find_class_invariant.sage"
load "igusa_invariants.sage"

